#!/usr/bin/env python3

import logging
from typing import BinaryIO, Dict, List, Any, Optional

from .mp3_boxes import (
    parse_id3v2_tag,
    search_for_image_data,
    get_mpeg_audio_properties,
    get_apic_frame_data,
)
from ..base import BaseMediaParser

logger = logging.getLogger(__name__)


class Mp3Parser(BaseMediaParser):
    def __init__(self):
        self.metadata: Dict[str, Any] = {}
        self.audio_tracks: List[Dict[str, Any]] = []
        self.total_file_size: int = 0
        self.id3_tag_size: int = 0

    def parse(self, f: BinaryIO) -> Dict[str, Any]:
        self.metadata = {
            "has_cover_art": None,
            "cover_art_mime": None,
            "cover_art_dimensions": None,
        }
        self.audio_tracks = []
        f.seek(0, 2)
        self.total_file_size = f.tell()
        f.seek(0)

        id3_tag_data = parse_id3v2_tag(f, self._apply_metadata_field)
        self.id3_tag_size = id3_tag_data.get("size", 0)
        if id3_tag_data.get("has_image"):
            self.metadata["has_cover_art"] = True
            self.metadata["cover_art_mime"] = id3_tag_data.get("image_mime")
            self.metadata["cover_art_dimensions"] = id3_tag_data.get("image_dimensions")

        if not self.metadata.get("has_cover_art"):
            image_data = search_for_image_data(
                f, self.id3_tag_size, self.total_file_size
            )
            if image_data:
                self.metadata.update(image_data)

        audio_info = get_mpeg_audio_properties(
            f, self.id3_tag_size, self.total_file_size
        )
        if audio_info:
            if (
                "duration_seconds" in self.metadata
                and self.metadata["duration_seconds"] is not None
            ):
                try:
                    duration_from_tag = int(self.metadata["duration_seconds"]) / 1000.0
                    audio_info["duration_seconds"] = duration_from_tag
                except (ValueError, TypeError):
                    logger.warning(
                        f"Could not convert ID3 TLEN '{self.metadata['duration_seconds']}' to seconds. Falling back to calculated duration."
                    )
                    audio_info["duration_seconds"] = None

            if (
                audio_info.get("duration_seconds") is None
                and audio_info.get("total_samples") is not None
                and audio_info.get("sample_rate") is not None
                and audio_info["sample_rate"] > 0
            ):
                audio_info["duration_seconds"] = (
                    audio_info["total_samples"] / audio_info["sample_rate"]
                )

            if audio_info.get("duration_seconds") is not None:
                self.metadata["duration_seconds"] = audio_info.get("duration_seconds")
            else:
                self.metadata.pop("duration_seconds", None)

            if (
                audio_info.get("duration_seconds") is not None
                and audio_info["duration_seconds"] > 0
            ):
                audio_data_size_bytes = self.total_file_size - self.id3_tag_size
                if audio_data_size_bytes > 0:
                    average_bitrate_bps = (audio_data_size_bytes * 8) / audio_info[
                        "duration_seconds"
                    ]
                    audio_info["bitrate"] = round(average_bitrate_bps)
                else:
                    audio_info["bitrate"] = None

            if (
                audio_info.get("bitrate") is None
                and audio_info.get("initial_frame_bitrate_kbps") is not None
            ):
                audio_info["bitrate"] = (
                    audio_info["initial_frame_bitrate_kbps"] * 1000
                )

            # Add the audio info to the tracks list.
            self.audio_tracks.append(
                {
                    "handler_name": "Audio",
                    "language": self.metadata.get("language", "und"),
                    **{
                        k: v
                        for k, v in audio_info.items()
                        if k not in ["initial_frame_bitrate_kbps"]
                    },
                }
            )

        final_audio_tracks = []
        for i, track in enumerate(self.audio_tracks):
            track["index"] = i
            ordered_track = self._order_audio_track(track)
            final_audio_tracks.append(ordered_track)

        total_bitrate = 0
        for track in final_audio_tracks:
            total_bitrate += track.get("bitrate") or 0

        if total_bitrate > 0:
            self.metadata["bitrate"] = total_bitrate

        final_metadata = self._process_metadata_for_output(self.metadata)

        return {
            "metadata": final_metadata,
            "audio": final_audio_tracks,
            "video": [],
            "subtitle": [],
        }

    def get_cover_art(self, f: BinaryIO) -> Optional[bytes]:
        """
        Extracts the raw cover art from the MP3 file's ID3v2 tag by finding
        the APIC frame.
        """
        f.seek(0)
        return get_apic_frame_data(f)

    def _order_audio_track(self, track: Dict[str, Any]) -> Dict[str, Any]:
        """Reorders audio track fields for consistent output."""
        ordered_keys = [
            "index",
            "handler_name",
            "language",
            "codec",
            "codec_tag_string",
            "channels",
            "channel_layout",
            "sample_rate",
            "bits_per_sample",
            "bitrate",
            "duration_seconds",
            "total_samples",
        ]

        ordered_dict = {}
        for key in ordered_keys:
            if key in track:
                ordered_dict[key] = track.pop(key)

        ordered_dict.update(track)
        return ordered_dict

    def _apply_metadata_field(self, key: str, value: Any):
        if key == "track_number":
            try:
                if isinstance(value, str) and "/" in value:
                    num_str, total_str = value.split("/", 1)
                    self.metadata[key] = int(num_str)
                    self.metadata["track_total"] = str(total_str)
                else:
                    self.metadata[key] = int(value)
            except (ValueError, TypeError):
                self.metadata[key] = value
        elif key == "disc_number":
            try:
                if isinstance(value, str) and "/" in value:
                    num_str, total_str = value.split("/", 1)
                    self.metadata[key] = int(num_str)
                    self.metadata["disc_total"] = str(total_str)
                else:
                    self.metadata[key] = int(value)
            except (ValueError, TypeError):
                self.metadata[key] = value
        elif key == "barcode":
            try:
                self.metadata[key] = int(value)
            except (ValueError, TypeError):
                self.metadata[key] = value
        elif key == "tempo":
            try:
                self.metadata[key] = int(float(value))
            except (ValueError, TypeError):
                self.metadata[key] = value
        elif key == "release_date":
            if isinstance(value, str):
                if len(value) == 4 and value.isdigit():
                    try:
                        year = int(value)
                        if 1900 < year < 2100:
                            self.metadata[key] = str(year)
                        else:
                            self.metadata[key] = value
                    except ValueError:
                        self.metadata[key] = value
                elif len(value) == 10 and value[4] == "-" and value[7] == "-":
                    self.metadata[key] = value
                else:
                    self.metadata[key] = value
            else:
                self.metadata[key] = value
        elif key in [
            "replaygain_track_gain",
            "replaygain_track_peak",
            "replaygain_album_gain",
            "replaygain_album_peak",
            "lyrics",
            "encoder",
            "comment",
            "copyright",
            "publisher",
            "performer",
            "language",
            "record_company",
            "upc",
            "media_type",
            "description",
            "isrc",
            "track_total",
            "disc_total",
        ]:
            self.metadata[key] = str(value).strip()
        elif key == "duration_seconds":
            try:
                self.metadata[key] = int(value)
            except (ValueError, TypeError):
                self.metadata[key] = value
        elif key == "itunesadvisory":
            if isinstance(value, bool):
                self.metadata[key] = "1" if value else "0"
            else:
                self.metadata[key] = str(value).strip()
        else:
            self.metadata[key] = value

    def _process_metadata_for_output(
        self, raw_metadata: Dict[str, Any]
    ) -> Dict[str, Any]:
        processed_metadata = {}
        output_order = [
            "title",
            "artist",
            "album",
            "album_artist",
            "track_number",
            "track_total",
            "disc_number",
            "disc_total",
            "genre",
            "duration_seconds",
            "bitrate",
            "release_date",
            "publisher",
            "isrc",
            "barcode",
            "upc",
            "media_type",
            "replaygain_track_gain",
            "replaygain_album_gain",
            "lyrics",
            "composer",
            "copyright",
            "has_cover_art",
            "cover_art_mime",
            "cover_art_dimensions",
            "comment",
            "encoder",
            "performer",
            "language",
            "record_company",
            "description",
            "tempo",
            "itunesadvisory",
        ]

        temp_data = {}
        for key in list(raw_metadata.keys()):
            temp_data[key] = raw_metadata.pop(key)

        for key_in_order in output_order:
            if key_in_order in temp_data:
                processed_metadata[key_in_order] = temp_data.pop(key_in_order)

        for key, value in temp_data.items():
            if (
                value is not None
                and value != ""
                and key
                not in [
                    "unique_file_identifier",
                    "tlen",
                    "tdat",
                ]
            ):
                processed_metadata[key] = value

        if "track_total" in processed_metadata:
            processed_metadata["track_total"] = int(processed_metadata["track_total"])
        if "disc_total" in processed_metadata:
            processed_metadata["disc_total"] = int(processed_metadata["disc_total"])
        if "itunesadvisory" in processed_metadata:
            try:
                processed_metadata["itunesadvisory"] = int(
                    processed_metadata["itunesadvisory"]
                )
            except (ValueError, TypeError):
                pass
        return processed_metadata



# metaspector/inspector.py
# !/usr/bin/env python3

import os
import logging
import struct
from typing import Dict, Optional, Any, BinaryIO

from .format_handlers.mp4.mp4 import Mp4Parser
from .format_handlers.flac.flac import FlacParser
from .format_handlers.mp3.mp3 import Mp3Parser

logger = logging.getLogger(__name__)


class MediaInspector:
    def __init__(self, filepath: str):
        if not os.path.exists(filepath):
            raise FileNotFoundError(f"File not found at '{filepath}'")
        self.filepath = filepath

    def _get_parser_instance(self, f: BinaryIO):
        """Internal helper to detect file type and return the correct parser instance."""
        initial_check_bytes = f.read(1024)
        f.seek(0)

        if initial_check_bytes.startswith(b"ID3"):
            return Mp3Parser()
        elif initial_check_bytes[0] == 0xFF and (initial_check_bytes[1] & 0xE0) == 0xE0:
            return Mp3Parser()
        elif initial_check_bytes[0:4] == b"fLaC":
            return FlacParser()

        # Robust MP4 check by searching for key atoms
        f.seek(0)
        pos = 0
        while pos < 1024:  # Search within the first 1KB
            f.seek(pos)
            size_bytes = f.read(4)
            if len(size_bytes) < 4:
                break
            size = struct.unpack(">I", size_bytes)[0]

            atom_type = f.read(4)
            if atom_type in (b"ftyp", b"moov", b"mdat"):
                f.seek(0)
                return Mp4Parser()

            if size == 0:
                break  # Cannot determine next position
            pos += size

        f.seek(0)
        return None

    def inspect(self, section: Optional[str] = None) -> Dict[str, Any]:
        """
        Inspects the media file and returns its metadata, audio, video, and subtitle tracks.
        """
        try:
            with open(self.filepath, "rb") as f:
                parser_instance = self._get_parser_instance(f)

                if parser_instance:
                    logger.info(f"Using parser: {type(parser_instance).__name__}")
                    result = parser_instance.parse(f)
                else:
                    raise ValueError(f"Unsupported file format for '{self.filepath}'.")

        except FileNotFoundError:
            raise
        except Exception as e:
            raise IOError(f"Error parsing media file '{self.filepath}': {e}") from e

        if section:
            if section in result:
                return {section: result[section]}
            else:
                raise ValueError(
                    f"Invalid section '{section}'. Available sections are: {', '.join(result.keys())}"
                )
        return result

    def get_cover_art(self) -> Optional[bytes]:
        """
        Extracts and returns the raw cover art bytes from the media file.
        """
        try:
            with open(self.filepath, "rb") as f:
                parser_instance = self._get_parser_instance(f)
                if parser_instance and hasattr(parser_instance, "get_cover_art"):
                    return parser_instance.get_cover_art(f)
                return None
        except Exception as e:
            logger.error(f"Error getting cover art: {e}")
            return None



# metaspector/format_handlers/mp3/mp3_boxes.py
# !/usr/bin/env python3

import struct
import logging
from typing import BinaryIO, Dict, Any, Callable, Optional

logger = logging.getLogger(__name__)

from .mp3_utils import (
    MPEG_VERSIONS,
    LAYER_DESCRIPTIONS,
    BITRATE_TABLES,
    SAMPLE_RATE_TABLES,
    CHANNEL_MODES,
    SAMPLES_PER_FRAME,
    ID3_KEY_MAP,
    parse_image_and_update_metadata,
    decode_id3_string,
)


def parse_id3v2_tag(f: BinaryIO, apply_metadata_func: Callable) -> Dict[str, Any]:
    """
    Parses the ID3v2 tag at the beginning of the file.
    """
    initial_pos = f.tell()
    header = f.read(10)
    if len(header) < 10 or header[0:3] != b"ID3":
        f.seek(initial_pos)
        return {"size": 0, "has_image": False}

    version_major = header[3]
    flags = header[5]
    size = (
        ((header[6] & 0x7F) << 21)
        | ((header[7] & 0x7F) << 14)
        | ((header[8] & 0x7F) << 7)
        | (header[9] & 0x7F)
    )
    id3_tag_size = 10 + size
    id3_tag_end = initial_pos + id3_tag_size

    if flags & 0x40:
        extended_header_start_pos = f.tell()
        extended_header_size_bytes = f.read(4)
        if len(extended_header_size_bytes) < 4:
            logger.warning("Incomplete extended header size. Skipping ID3 parsing.")
            f.seek(id3_tag_end)
            return {"size": id3_tag_size, "has_image": False}
        extended_header_size = struct.unpack(">I", extended_header_size_bytes)[0]
        if (
            extended_header_size > 4
            and extended_header_start_pos + extended_header_size <= id3_tag_end
        ):
            f.read(extended_header_size - 4)
        else:
            logger.warning(
                f"Malformed extended header size {extended_header_size}. Skipping."
            )
            f.seek(id3_tag_end)
            return {"size": id3_tag_size, "has_image": False}

    has_image = False
    image_mime = None
    image_dimensions = None

    while f.tell() < id3_tag_end:
        frame_header = f.read(10)
        if len(frame_header) < 10:
            logger.warning("Incomplete ID3v2 frame header. Stopping ID3 parsing.")
            break

        frame_id = frame_header[0:4].decode("ascii", errors="ignore").strip("\x00")
        if all(b == 0 for b in frame_header):
            break

        if version_major == 4:
            frame_size = (
                ((frame_header[4] & 0x7F) << 21)
                | ((frame_header[5] & 0x7F) << 14)
                | ((frame_header[6] & 0x7F) << 7)
                | (frame_header[7] & 0x7F)
            )
        else:
            frame_size = struct.unpack(">I", frame_header[4:8])[0]

        remaining_in_tag = id3_tag_end - f.tell()
        if frame_size <= 0 or frame_size > remaining_in_tag:
            logger.warning(
                f"Invalid or truncated ID3v2 frame '{frame_id}' (size {frame_size}, remaining {remaining_in_tag}). Skipping."
            )
            f.seek(id3_tag_end)
            break

        frame_data = f.read(frame_size)
        if len(frame_data) < frame_size:
            logger.warning(
                f"Premature EOF while reading data for frame '{frame_id}'. Expected {frame_size}, read {len(frame_data)}. Stopping ID3 parsing."
            )
            break

        parsed_frame = _parse_id3v2_frame_content(frame_id, frame_data)
        if parsed_frame:
            if frame_id == "APIC" and parsed_frame.get("has_cover_art"):
                has_image = True
                image_mime = parsed_frame.get("cover_art_mime")
                image_dimensions = parsed_frame.get("cover_art_dimensions")
            elif frame_id != "APIC":
                for key, value in parsed_frame.items():
                    apply_metadata_func(key, value)
    f.seek(id3_tag_end)
    return {
        "size": id3_tag_size,
        "has_image": has_image,
        "image_mime": image_mime,
        "image_dimensions": image_dimensions,
    }


def _parse_id3v2_frame_content(frame_id: str, data: bytes) -> Dict[str, Any]:
    """
    Parses the content of a single ID3v2 frame.
    """
    metadata = {}
    if frame_id.startswith("T") and frame_id != "TXXX" and frame_id != "TFLT":
        if not data:
            return {}
        encoding = data[0]
        content = data[1:]
        decoded_content = decode_id3_string(content, encoding)
        normalized_key = ID3_KEY_MAP.get(frame_id, frame_id.lower())
        metadata[normalized_key] = decoded_content
    elif frame_id == "USLT":
        if not data or len(data) < 7:
            return {}
        encoding = data[0]
        language = data[1:4].decode("ascii", errors="ignore").strip("\x00")
        null_term_size = 2 if encoding == 0x01 or encoding == 0x02 else 1
        desc_start_offset = 4
        desc_end_index = -1
        if null_term_size == 2:
            for i in range(
                desc_start_offset, len(data) - null_term_size + 1, null_term_size
            ):
                if data[i : i + null_term_size] == b"\x00" * null_term_size:
                    desc_end_index = i
                    break
        else:
            desc_end_index = data.find(b"\x00", desc_start_offset)
        if desc_end_index == -1:
            text_content = data[desc_start_offset:]
        else:
            text_content = data[desc_end_index + null_term_size :]
        decoded_lyrics = decode_id3_string(text_content, encoding)
        normalized_lyrics = (
            decoded_lyrics.replace("\r\n", "\n").replace("\r", "\n").strip()
        )
        metadata["lyrics"] = normalized_lyrics
        metadata["language"] = language
    elif frame_id == "TXXX":
        if not data:
            return {}
        encoding = data[0]
        content = data[1:]
        null_term_size = 2 if encoding == 0x01 or encoding == 0x02 else 1
        key_end_index = -1
        if null_term_size == 2:
            for i in range(0, len(content) - null_term_size + 1, null_term_size):
                if content[i : i + null_term_size] == b"\x00" * null_term_size:
                    key_end_index = i
                    break
        else:
            key_end_index = content.find(b"\x00")
        if key_end_index == -1:
            logger.warning("TXXX frame: missing key null terminator.")
            return {}
        key_raw = content[:key_end_index]
        value_raw = content[key_end_index + null_term_size :]
        key = decode_id3_string(key_raw, encoding)
        value = decode_id3_string(value_raw, encoding)
        normalized_key = ID3_KEY_MAP.get(f"TXXX:{key.upper()}", key.lower())
        if key.upper() == "CM/REPUBLIC":
            normalized_key = "publisher"
        elif key.upper() == "TSSE":
            normalized_key = "encoder"
        metadata[normalized_key] = value
    elif frame_id == "APIC":
        metadata = {
            "has_cover_art": False,
            "cover_art_mime": None,
            "cover_art_dimensions": None,
        }
        if not data or len(data) < 2:
            return metadata
        offset = 0
        encoding = data[offset]
        offset += 1
        null_term = b"\x00"
        try:
            mime_end = data.index(null_term, offset)
            mime_type = data[offset:mime_end].decode("latin-1", errors="ignore")
            offset = mime_end + 1
        except ValueError:
            return metadata
        if len(data) <= offset:
            return metadata
        offset += 1
        if encoding in [0x01, 0x02]:
            null_term = b"\x00\x00"
        else:
            null_term = b"\x00"
        try:
            desc_end = data.index(null_term, offset)
            offset = desc_end + len(null_term)
        except ValueError:
            pass
        image_data = data[offset:]
        if image_data:
            image_info = parse_image_and_update_metadata(image_data, mime_type)
            metadata.update(image_info)
    elif frame_id == "UFID":
        if not data or b"\x00" not in data:
            return {}
        owner_id_end = data.find(b"\x00", 1)
        if owner_id_end == -1:
            return {}
        owner_id = data[1:owner_id_end].decode("latin-1", errors="replace").strip()
        identifier_data = data[owner_id_end + 1 :]
        if owner_id == "http://www.id3.org/uslt/iTunes":
            try:
                if identifier_data.startswith(b"isrc"):
                    isrc = (
                        identifier_data[4:]
                        .decode("latin-1", errors="replace")
                        .strip("\x00")
                    )
                    metadata["isrc"] = isrc
                elif identifier_data.startswith(b"barcode"):
                    barcode = (
                        identifier_data[7:]
                        .decode("latin-1", errors="replace")
                        .strip("\x00")
                    )
                    metadata["barcode"] = barcode
            except UnicodeDecodeError:
                pass
    return metadata


def search_for_image_data(
    f: BinaryIO, id3_tag_size: int, total_file_size: int
) -> Optional[Dict[str, Any]]:
    """
    Searches for an image embedded in the audio data if not found in the ID3 tag.
    """
    initial_pos = f.tell()
    f.seek(id3_tag_size)
    data_chunk_size = 64 * 1024

    while f.tell() < total_file_size:
        start_pos = f.tell()
        chunk = f.read(data_chunk_size)
        if not chunk:
            break

        jpeg_start = chunk.find(b"\xff\xd8\xff")
        if jpeg_start != -1:
            f.seek(start_pos + jpeg_start)
            image_data = f.read()
            f.seek(initial_pos)
            return parse_image_and_update_metadata(image_data, "image/jpeg")

        png_start = chunk.find(b"\x89PNG\r\n\x1a\n")
        if png_start != -1:
            f.seek(start_pos + png_start)
            image_data = f.read()
            f.seek(initial_pos)
            return parse_image_and_update_metadata(image_data, "image/png")

    f.seek(initial_pos)
    return None


def get_mpeg_audio_properties(
    f: BinaryIO, id3_tag_size: int, total_file_size: int
) -> Optional[Dict[str, Any]]:
    """
    Searches for and parses the first valid MPEG audio frame to get properties.
    """
    original_pos = f.tell()
    f.seek(id3_tag_size)
    initial_audio_data_pos = f.tell()
    buffer_size = 64 * 1024
    frame_length_bytes = None
    first_frame_bitrate = None
    first_frame_sample_rate = None
    first_frame_channels = None
    first_frame_channel_layout = None  # Added for channel layout
    first_frame_codec = None
    first_frame_codec_tag = None
    first_frame_samples_per_frame = None

    while True:
        current_read_pos = f.tell()
        data_chunk = f.read(buffer_size)
        if not data_chunk:
            break
        for i in range(len(data_chunk) - 4):
            header_bytes = data_chunk[i : i + 4]
            if header_bytes[0] == 0xFF and (header_bytes[1] & 0xE0) == 0xE0:
                try:
                    header_int = struct.unpack(">I", header_bytes)[0]
                except struct.error:
                    continue
                mpeg_version_bits = (header_int >> 19) & 0x03
                layer_bits = (header_int >> 17) & 0x03
                bitrate_index = (header_int >> 12) & 0x0F
                sample_rate_index = (header_int >> 10) & 0x03
                padding_bit = (header_int >> 9) & 0x01
                channel_mode_bits = (header_int >> 6) & 0x03
                if (
                    mpeg_version_bits == 0b01
                    or layer_bits == 0b00
                    or bitrate_index == 0
                    or bitrate_index == 0x0F
                    or sample_rate_index == 0x03
                ):
                    continue
                mpeg_version = MPEG_VERSIONS.get(mpeg_version_bits)
                layer = LAYER_DESCRIPTIONS.get(layer_bits)
                channels_str = CHANNEL_MODES.get(channel_mode_bits)
                if not mpeg_version or not layer or channels_str is None:
                    continue
                bitrate_map_key = f"{mpeg_version}_{layer.replace(' ', '-')}"
                bitrate_kbps_from_frame = BITRATE_TABLES.get(bitrate_map_key, {}).get(
                    bitrate_index
                )
                sample_rate = SAMPLE_RATE_TABLES.get(mpeg_version, {}).get(
                    sample_rate_index
                )
                if bitrate_kbps_from_frame is None or sample_rate is None:
                    continue

                # Determine channels and channel layout from channel mode
                channels = 1 if channels_str == "Mono" else 2
                if channels_str == "Stereo" or channels_str == "Joint Stereo":
                    channel_layout = "2.0"
                elif channels_str == "Dual Channel":
                    channel_layout = "1+1"
                elif channels_str == "Mono":
                    channel_layout = "1.0"
                else:
                    channel_layout = None

                sps_key = f"{mpeg_version}_{layer.replace(' ', '-')}"
                samples_per_frame = SAMPLES_PER_FRAME.get(sps_key)
                if not samples_per_frame:
                    continue
                if layer_bits == 0b11:
                    frame_length_bytes = (
                        int(
                            ((12 * bitrate_kbps_from_frame * 1000.0) / sample_rate)
                            + padding_bit
                        )
                        * 4
                    )
                else:
                    frame_length_bytes = int(
                        ((144 * bitrate_kbps_from_frame * 1000.0) / sample_rate)
                        + padding_bit
                    )
                if frame_length_bytes <= 0:
                    continue
                check_pos = current_read_pos + i + frame_length_bytes
                if check_pos + 4 <= total_file_size:
                    f.seek(check_pos)
                    next_header_peek = f.read(4)
                    f.seek(current_read_pos + i)
                    if (
                        len(next_header_peek) == 4
                        and next_header_peek[0] == 0xFF
                        and (next_header_peek[1] & 0xE0) == 0xE0
                    ):
                        first_frame_bitrate = bitrate_kbps_from_frame
                        first_frame_sample_rate = sample_rate
                        first_frame_channels = channels
                        first_frame_channel_layout = channel_layout  # Store layout

                        if layer == "Layer III":
                            first_frame_codec = "mp3"
                            first_frame_codec_tag = "mp3"
                        elif layer == "Layer II":
                            first_frame_codec = "mp2"
                            first_frame_codec_tag = "mp2"
                        else:
                            first_frame_codec = "mp1"
                            first_frame_codec_tag = "mp1"

                        first_frame_samples_per_frame = samples_per_frame
                        initial_audio_data_pos = current_read_pos + i
                        break
        if first_frame_bitrate is not None:
            break
        f.seek(current_read_pos + len(data_chunk) - 4)
        if f.tell() >= total_file_size:
            break
    f.seek(original_pos)
    if first_frame_bitrate is None:
        return None
    total_samples = 0
    if (
        first_frame_sample_rate > 0
        and first_frame_bitrate > 0
        and total_file_size > initial_audio_data_pos
    ):
        estimated_audio_bytes = total_file_size - initial_audio_data_pos
        if frame_length_bytes > 0:
            estimated_total_frames = estimated_audio_bytes / frame_length_bytes
            total_samples = int(estimated_total_frames * first_frame_samples_per_frame)
    return {
        "codec": first_frame_codec,
        "codec_tag_string": first_frame_codec_tag,
        "channels": first_frame_channels,
        "channel_layout": first_frame_channel_layout,  # Add to return dict
        "sample_rate": first_frame_sample_rate,
        "bits_per_sample": 16,
        "duration_seconds": None,
        "total_samples": total_samples,
        "initial_frame_bitrate_kbps": first_frame_bitrate,
    }


def get_apic_frame_data(f: BinaryIO) -> Optional[bytes]:
    """
    Specifically searches for an APIC frame in an ID3v2 tag and extracts the raw image data.
    This function is self-contained and safe to call for cover art extraction.
    """
    f.seek(0)
    header = f.read(10)
    if len(header) < 10 or header[0:3] != b"ID3":
        return None

    version_major = header[3]
    size = (
        ((header[6] & 0x7F) << 21)
        | ((header[7] & 0x7F) << 14)
        | ((header[8] & 0x7F) << 7)
        | (header[9] & 0x7F)
    )
    id3_tag_end = 10 + size

    if header[5] & 0x40:
        f.seek(10)
        ext_header_size_bytes = f.read(4)
        if len(ext_header_size_bytes) < 4:
            return None
        ext_header_size = struct.unpack(">I", ext_header_size_bytes)[0]
        if ext_header_size > 4:
            f.read(ext_header_size - 4)

    while f.tell() < id3_tag_end:
        frame_header = f.read(10)
        if len(frame_header) < 10 or all(b == 0 for b in frame_header):
            break

        frame_id = frame_header[0:4].decode("ascii", errors="ignore")

        if version_major == 4:
            frame_size = (
                ((frame_header[4] & 0x7F) << 21)
                | ((frame_header[5] & 0x7F) << 14)
                | ((frame_header[6] & 0x7F) << 7)
                | (frame_header[7] & 0x7F)
            )
        else:
            frame_size = struct.unpack(">I", frame_header[4:8])[0]

        current_pos = f.tell()
        if frame_size <= 0 or current_pos + frame_size > id3_tag_end:
            break

        if frame_id == "APIC":
            frame_data = f.read(frame_size)
            try:
                encoding_byte = frame_data[0]
                offset = 1  # Start after encoding byte

                # Skip over MIME type string
                mime_end = frame_data.index(b"\x00", offset)
                offset = mime_end + 1

                # Skip picture type byte
                offset += 1

                # Skip over description string
                null_term = b"\x00\x00" if encoding_byte in [1, 2] else b"\x00"
                desc_end = frame_data.index(null_term, offset)
                offset = desc_end + len(null_term)

                # The rest of the frame is the image data
                return frame_data[offset:]
            except (ValueError, IndexError):
                # Malformed APIC frame, continue searching
                f.seek(current_pos + frame_size)
                continue
        else:
            f.seek(frame_size, 1)

    return None



# metaspector/format_handlers/mp4/mp4_bitstream_parser.py
# !/usr/bin/env python3

import logging
import struct

from typing import BinaryIO, Dict, Any, Optional
from .mp4_utils import (
    _read_box_header,
    _read_uint8,
    _read_uint32,
    _read_uint64,
    COLOR_PRIMARIES_MAP,
    TRANSFER_CHARACTERISTICS_MAP,
    MATRIX_COEFFICIENTS_MAP,
)

logger = logging.getLogger(__name__)


class BitReader:
    """A helper class for reading bits from a byte stream."""

    def __init__(self, data: bytes):
        self.data = data
        self.byte_pos = 0
        self.bit_pos = 0

    def read_bit(self) -> int:
        if self.byte_pos >= len(self.data):
            raise IndexError("Reading past end of data")
        byte = self.data[self.byte_pos]
        bit = (byte >> (7 - self.bit_pos)) & 1
        self.bit_pos += 1
        if self.bit_pos == 8:
            self.bit_pos = 0
            self.byte_pos += 1
        return bit

    def read_bits(self, num_bits: int) -> int:
        value = 0
        for _ in range(num_bits):
            value = (value << 1) | self.read_bit()
        return value

    def read_ue(self) -> int:
        leading_zeros = 0
        while self.byte_pos < len(self.data) and self.read_bit() == 0:
            leading_zeros += 1
            if self.byte_pos >= len(self.data) and self.bit_pos == 0:
                break

        if self.byte_pos >= len(self.data) and leading_zeros > 0:
            raise IndexError("Reading past end of data while parsing Exp-Golomb")

        return (1 << leading_zeros) - 1 + self.read_bits(leading_zeros)

    def read_se(self) -> int:
        value = self.read_ue()
        if value & 1:
            return (value + 1) // 2
        else:
            return -1 * (value // 2)

    def byte_aligned(self) -> bool:
        return self.bit_pos == 0

    def skip_to_next_byte(self):
        if self.bit_pos != 0:
            self.bit_pos = 0
            self.byte_pos += 1


class BitstreamParser:
    """
    A collection of static methods for parsing bitstream-level data
    (e.g., video SPS/VUI, SEI messages) and extracting sample data.
    """

    @staticmethod
    def extract_sample_data(
        f: BinaryIO, moov_start: int, moov_end: int, index: int
    ) -> Optional[bytes]:
        """
        Navigates the MP4 structure to find and extract the raw data of the first
        sample for a given index. This is essential for bitstream parsing.
        """
        logger.debug(f"Attempting to extract first sample for track ID: {index}")
        f.seek(moov_start)

        while f.tell() < moov_end:
            box_type, _, box_start, box_end = _read_box_header(f)
            if not box_type:
                break
            if f.tell() > moov_end:
                break

            if box_type == b"trak":
                trak_start = f.tell() - 8

                is_correct_track = False
                f.seek(trak_start)
                while f.tell() < box_end:
                    inner_type, _, _, inner_end = _read_box_header(f)
                    if not inner_type or inner_end > box_end:
                        break
                    if inner_type == b"tkhd":
                        f.read(8)  # version + flags
                        version = _read_uint8(f)
                        f.read(3)  # flags
                        if version == 1:
                            f.read(16)
                            found_id = _read_uint32(f)
                        elif version == 0:
                            f.read(8)
                            found_id = _read_uint32(f)
                        else:
                            found_id = None

                        if found_id == index:
                            is_correct_track = True
                        break
                    f.seek(inner_end)

                if is_correct_track:
                    logger.debug(f"Found correct 'trak' for track ID {index}")
                    f.seek(trak_start)
                    stbl_start, stbl_end = 0, 0
                    current_pos_trak = trak_start
                    while current_pos_trak < box_end:
                        f.seek(current_pos_trak)
                        b_type, _, b_start, b_end = _read_box_header(f)
                        if not b_type or b_end > box_end:
                            break
                        if b_type == b"stbl":
                            stbl_start, stbl_end = b_start, b_end
                            break
                        if b_type in (b"mdia", b"minf"):
                            current_pos_trak = b_start
                        else:
                            current_pos_trak = b_end

                    if not stbl_start:
                        logger.error("Could not find 'stbl' box for the track.")
                        return None

                    first_sample_size = None
                    first_chunk_offset = None
                    f.seek(stbl_start)
                    while f.tell() < stbl_end:
                        stbl_box_type, _, stbl_box_start, stbl_box_end = (
                            _read_box_header(f)
                        )
                        if not stbl_box_type or stbl_box_end > stbl_end:
                            break

                        if stbl_box_type == b"stsz":
                            f.seek(stbl_box_start + 8)
                            sample_size = _read_uint32(f)
                            entry_count = _read_uint32(f)
                            if sample_size == 0 and entry_count > 0:
                                first_sample_size = _read_uint32(f)
                            else:
                                first_sample_size = sample_size
                            logger.debug(
                                f"Found first sample size: {first_sample_size}"
                            )

                        elif stbl_box_type == b"stco":
                            f.seek(stbl_box_start + 8)
                            entry_count = _read_uint32(f)
                            if entry_count > 0:
                                first_chunk_offset = _read_uint32(f)
                            logger.debug(
                                f"Found 32-bit chunk offset: {first_chunk_offset}"
                            )

                        elif stbl_box_type == b"co64":
                            f.seek(stbl_box_start + 8)
                            entry_count = _read_uint32(f)
                            if entry_count > 0:
                                first_chunk_offset = _read_uint64(f)
                            logger.debug(
                                f"Found 64-bit chunk offset: {first_chunk_offset}"
                            )

                        if (
                            first_sample_size is not None
                            and first_chunk_offset is not None
                        ):
                            break

                        f.seek(stbl_box_end)

                    if first_sample_size is not None and first_chunk_offset is not None:
                        try:
                            logger.info(
                                f"Extracting sample of size {first_sample_size} from offset {first_chunk_offset}"
                            )
                            f.seek(first_chunk_offset)
                            return f.read(first_sample_size)
                        except Exception as e:
                            logger.error(f"Failed to seek/read sample data: {e}")
                            return None

            f.seek(box_end)

        logger.warning(
            f"Could not find sample data for track ID {index}. 'trak' may not exist or parsing failed."
        )
        return None

    @staticmethod
    def _parse_vui_parameters(reader: BitReader, details: Dict[str, Any]):
        """Parses VUI parameters from a video bitstream."""
        try:
            if reader.read_bit():  # aspect_ratio_info_present_flag
                aspect_ratio_idc = reader.read_bits(8)
                if aspect_ratio_idc == 255:  # SAR_Extended
                    reader.read_bits(16)  # sar_width
                    reader.read_bits(16)  # sar_height

            if reader.read_bit():  # overscan_info_present_flag
                reader.read_bit()  # overscan_appropriate_flag

            if reader.read_bit():  # video_signal_type_present_flag
                reader.read_bits(3)  # video_format
                details["video_full_range_flag"] = reader.read_bit()
                if reader.read_bit():  # colour_description_present_flag
                    details["color_primaries"] = reader.read_bits(8)
                    details["transfer_characteristics"] = reader.read_bits(8)
                    details["matrix_coefficients"] = reader.read_bits(8)
                    logger.debug(
                        f"VUI: Primaries: {details['color_primaries']}, Transfer: {details['transfer_characteristics']}, Matrix: {details['matrix_coefficients']}, Range: {details.get('video_full_range_flag')}"
                    )

            if reader.read_bit():  # chroma_loc_info_present_flag
                reader.read_ue()  # chroma_sample_loc_type_top_field
                reader.read_ue()  # chroma_sample_loc_type_bottom_field

            if reader.read_bit():  # neutral_chroma_indication_flag
                pass
            if reader.read_bit():  # field_seq_flag
                pass
            if reader.read_bit():  # frame_field_info_present_flag
                pass

            # Simplified parsing for the rest of VUI
        except IndexError:
            logger.warning("Reached end of VUI payload prematurely during parsing.")
        except Exception as e:
            logger.error(f"Error parsing VUI parameters: {e}")

    @staticmethod
    def _parse_sps_payload(nal_payload: bytes, details: Dict[str, Any]):
        """
        Parses a SPS NAL unit payload to extract VUI parameters.
        This is a simplified implementation focusing only on VUI for color info.
        """
        reader = BitReader(nal_payload)

        try:
            reader.read_bits(4)
            reader.read_bits(3)
            reader.read_bit()

            reader.read_bits(2)
            reader.read_bit()
            reader.read_bits(5)

            reader.read_bits(32)
            reader.read_bits(48)

            reader.read_bits(8)

            sps_max_sub_layers_minus1 = details.get("sps_max_sub_layers_minus1", 0)
            for i in range(sps_max_sub_layers_minus1 + 1):
                pass

            reader.read_ue()

            chroma_format_idc = reader.read_ue()

            if chroma_format_idc == 3:
                reader.read_bit()

            reader.read_ue()
            reader.read_ue()

            if reader.read_bit():
                reader.read_ue()
                reader.read_ue()
                reader.read_ue()
                reader.read_ue()

            reader.read_ue()
            reader.read_ue()

            reader.read_ue()

            sps_sub_layer_ordering_info_present_flag = reader.read_bit()
            for i in range(sps_max_sub_layers_minus1 + 1):
                reader.read_ue()
                reader.read_ue()
                reader.read_ue()
                if not sps_sub_layer_ordering_info_present_flag:
                    break

            reader.read_ue()
            reader.read_ue()
            reader.read_ue()
            reader.read_ue()
            reader.read_ue()
            reader.read_ue()

            if reader.read_bit():
                if reader.read_bit():
                    pass

            reader.read_bit()
            reader.read_bit()
            reader.read_bit()
            if reader.read_bit():
                reader.read_bits(4)
                reader.read_bits(4)
                reader.read_ue()
                reader.read_ue()
                reader.read_bit()

            reader.read_ue()

            reader.read_bit()
            if reader.read_bit():
                reader.read_ue()

            reader.read_bit()
            reader.read_bit()

            if reader.read_bit():
                logger.debug("VUI parameters present in SPS.")
                BitstreamParser._parse_vui_parameters(reader, details)

            if reader.read_bit():
                pass

            if not reader.byte_aligned():
                reader.skip_to_next_byte()

        except IndexError:
            logger.warning("Reached end of SPS NAL payload prematurely during parsing.")
        except Exception as e:
            logger.error(f"Error parsing SPS NAL unit: {e}")

    @staticmethod
    def parse_video_bitstream(data: bytes) -> Dict[str, Any]:
        """
        Parses a video bitstream for SPS (VUI) and SEI messages containing HDR metadata.
        """
        bitstream_details = {
            "hdr_format": "SDR",
            "color_primaries": "Unknown",
            "transfer_characteristics": "Unknown",
            "matrix_coefficients": "Unknown",
            "color_space": "Unknown",
            "color_transfer": "Unknown",
            "color_range": "Unknown",
            "max_content_light_level": None,
            "max_frame_average_light_level": None,
            "dolby_vision_profile": None,
            "dolby_vision_level": None,
        }

        pos = 0
        logger.debug(f"Parsing video bitstream from {len(data)} bytes")
        while pos < len(data):
            start_code_pos = -1
            start_code_len = 0

            if pos + 4 <= len(data) and data[pos : pos + 4] == b"\x00\x00\x00\x01":
                start_code_pos = pos
                start_code_len = 4
            elif pos + 3 <= len(data) and data[pos : pos + 3] == b"\x00\x00\x01":
                start_code_pos = pos
                start_code_len = 3

            if start_code_pos == -1:
                pos += 1
                continue

            pos = start_code_pos + start_code_len
            if pos + 2 > len(data):
                break

            nal_unit_type = (data[pos] >> 1) & 0x3F

            next_start_code_pos = len(data)
            temp_pos = pos + 2
            while temp_pos < len(data) - 3:
                if data[temp_pos : temp_pos + 4] == b"\x00\x00\x00\x01":
                    next_start_code_pos = temp_pos
                    break
                elif data[temp_pos : temp_pos + 3] == b"\x00\x00\x01":
                    next_start_code_pos = temp_pos
                    break
                temp_pos += 1

            nal_end = next_start_code_pos
            nal_payload = data[pos + 2 : nal_end]

            logger.debug(
                f"Found NAL unit type {nal_unit_type} at offset {start_code_pos}. Payload size: {len(nal_payload)}"
            )

            if nal_unit_type == 33:
                logger.debug("Parsing SPS NAL unit.")
                BitstreamParser._parse_sps_payload(nal_payload, bitstream_details)

                if isinstance(bitstream_details.get("color_primaries"), int):
                    bitstream_details["color_primaries"] = COLOR_PRIMARIES_MAP.get(
                        bitstream_details["color_primaries"],
                        f"Unknown ({bitstream_details['color_primaries']})",
                    )

                if isinstance(bitstream_details.get("transfer_characteristics"), int):
                    bitstream_details["transfer_characteristics"] = (
                        TRANSFER_CHARACTERISTICS_MAP.get(
                            bitstream_details["transfer_characteristics"],
                            f"Unknown ({bitstream_details['transfer_characteristics']})",
                        )
                    )

                if isinstance(bitstream_details.get("matrix_coefficients"), int):
                    bitstream_details["matrix_coefficients"] = (
                        MATRIX_COEFFICIENTS_MAP.get(
                            bitstream_details["matrix_coefficients"],
                            f"Unknown ({bitstream_details['matrix_coefficients']})",
                        )
                    )

                if isinstance(bitstream_details.get("video_full_range_flag"), int):
                    bitstream_details["color_range"] = (
                        "full"
                        if bitstream_details["video_full_range_flag"] == 1
                        else "tv"
                    )

                if bitstream_details.get("matrix_coefficients") != "Unknown":
                    bitstream_details["color_space"] = bitstream_details[
                        "matrix_coefficients"
                    ]
                if bitstream_details.get("transfer_characteristics") != "Unknown":
                    bitstream_details["color_transfer"] = bitstream_details[
                        "transfer_characteristics"
                    ]

            elif nal_unit_type in (39, 40):
                logger.debug(f"Parsing SEI NAL unit (type {nal_unit_type}).")
                sei_pos = 0
                while sei_pos < len(nal_payload):
                    payload_type = 0
                    while sei_pos < len(nal_payload) and nal_payload[sei_pos] == 0xFF:
                        payload_type += 255
                        sei_pos += 1
                    if sei_pos >= len(nal_payload):
                        break
                    payload_type += nal_payload[sei_pos]
                    sei_pos += 1

                    payload_size = 0
                    while sei_pos < len(nal_payload) and nal_payload[sei_pos] == 0xFF:
                        payload_size += 255
                        sei_pos += 1
                    if sei_pos >= len(nal_payload):
                        break
                    payload_size += nal_payload[sei_pos]
                    sei_pos += 1

                    if sei_pos + payload_size > len(nal_payload):
                        logger.warning(
                            f"Invalid SEI payload size {payload_size} for type {payload_type} at sei_pos {sei_pos}. Skipping rest of NAL."
                        )
                        break

                    current_payload_data = nal_payload[sei_pos : sei_pos + payload_size]
                    logger.debug(
                        f"Processing SEI payload type {payload_type}, size {payload_size}"
                    )

                    try:
                        if payload_type == 137:
                            if len(current_payload_data) >= 24:
                                bitstream_details["transfer_characteristics"] = (
                                    "smpte2084"
                                )
                                if (
                                    bitstream_details["color_primaries"] == "Unknown"
                                    or bitstream_details["color_primaries"] == "bt709"
                                ):
                                    bitstream_details["color_primaries"] = "bt2020"
                                logger.debug(
                                    "Found Mastering Display Colour Volume (HDR10) SEI."
                                )
                        elif payload_type == 144:
                            if len(current_payload_data) >= 4:
                                max_cll, max_fall = struct.unpack(
                                    ">HH", current_payload_data[:4]
                                )
                                bitstream_details["max_content_light_level"] = max_cll
                                bitstream_details["max_frame_average_light_level"] = (
                                    max_fall
                                )
                                logger.debug(
                                    f"Found Content Light Level SEI: MaxCLL={max_cll}, MaxFALL={max_fall}"
                                )
                        elif payload_type == 5:
                            if (
                                len(current_payload_data) > 16
                                and current_payload_data[:12]
                                == b"\x44\x4f\x56\x49\x03\x01\x01\x08\x00\x00\x00\x00"
                            ):
                                logger.debug(
                                    f"Raw Dolby Vision data (payload_type 5): {current_payload_data.hex()}"
                                )
                                if len(current_payload_data) >= 27:
                                    val = current_payload_data[25]
                                    bitstream_details["dolby_vision_profile"] = (
                                        val >> 1
                                    ) & 0x7F
                                    bitstream_details["dolby_vision_level"] = (
                                        current_payload_data[26] & 0x3F
                                    )
                                    bitstream_details["hdr_format"] = "Dolby Vision"
                                    logger.debug(
                                        f"Found Dolby Vision SEI (heuristic): profile={bitstream_details['dolby_vision_profile']}, level={bitstream_details['dolby_vision_level']}"
                                    )
                        elif payload_type == 147:
                            if len(current_payload_data) >= 1:
                                transfer = current_payload_data[0]
                                if transfer == 18:
                                    bitstream_details["transfer_characteristics"] = (
                                        "arib-std-b67"
                                    )
                                    bitstream_details["hdr_format"] = "HLG"
                                    logger.debug(
                                        "Found Alternative Transfer Characteristics (HLG) SEI."
                                    )
                    except Exception as e:
                        logger.error(
                            f"Error parsing SEI payload type {payload_type}: {e}"
                        )

                    sei_pos += payload_size
            pos = nal_end

        if bitstream_details["dolby_vision_profile"] is not None:
            bitstream_details["hdr_format"] = "Dolby Vision"
            if bitstream_details["color_primaries"] == "Unknown":
                bitstream_details["color_primaries"] = "bt2020"
            if bitstream_details["transfer_characteristics"] == "Unknown":
                bitstream_details["transfer_characteristics"] = "smpte2084"
            if bitstream_details["matrix_coefficients"] == "Unknown":
                bitstream_details["matrix_coefficients"] = "bt2020nc"

        elif bitstream_details["transfer_characteristics"] == "arib-std-b67":
            bitstream_details["hdr_format"] = "HLG"
        elif bitstream_details["transfer_characteristics"] == "smpte2084":
            bitstream_details["hdr_format"] = "HDR (PQ)"

        logger.debug(f"Bitstream parsing result: {bitstream_details}")
        return bitstream_details



# metaspector/matrices/language_matrix.py
# !/usr/bin/env python3

from typing import List, Tuple, Optional


def get_language_matrix() -> List[Tuple[str, ...]]:
    """Returns the comprehensive language matrices."""
    language_matrix_references = [
        ("aa-ET", "aa", "aar", "Afar", "Qafar af", "Ethiopia"),
        ("ab-GE", "ab", "abk", "Abkhazian", "аҧсуа бызшәа", "Georgia"),
        ("ae-IR", "ae", "ave", "Avestan", "Avesta", "Iran"),
        ("af-ZA", "af", "afr", "Afrikaans", "Afrikaans", "South Africa"),
        ("ak-GH", "ak", "aka", "Akan", "Akan", "Ghana"),
        ("am-ET", "am", "amh", "Amharic", "አማርኛ (Amarəñña)", "Ethiopia"),
        ("an-ES", "an", "arg", "Aragonese", "Aragonés", "Spain"),
        ("ar-SA", "ar", "ara", "Arabic", "العربية", "Saudi Arabia"),
        ("as-IN", "as", "asm", "Assamese", "অসমীয়া (Ôxômiya)", "India"),
        ("av-RU", "av", "ava", "Avaric", "Авар мацӏ (Avar maⱬ)", "Russian Federation"),
        ("ay-BO", "ay", "aym", "Aymara", "Aymar aru", "Bolivia"),
        ("az-AZ", "az", "aze", "Azerbaijani", "Azərbaycan dili", "Azerbaijan"),
        ("ba-RU", "ba", "bak", "Bashkir", "Башҡорт теле", "Russian Federation"),
        ("be-BY", "be", "bel", "Belarusian", "беларуская мова", "Belarus"),
        ("bg-BG", "bg", "bul", "Bulgarian", "български език", "Bulgaria"),
        ("bi-VU", "bi", "bis", "Bislama", "Bislama", "Vanuatu"),
        ("bm-ML", "bm", "bam", "Bambara", "Bamanankan", "Mali"),
        ("bn-BD", "bn", "ben", "Bengali", "বাংলা (Bangla)", "Bangladesh"),
        ("bo-CN", "bo", "bod", "Tibetan", "བོད་ཡིག།", "China"),
        ("br-FR", "br", "bre", "Breton", "Brezhoneg", "France"),
        ("bs-BA", "bs", "bos", "Bosnian", "Bosanski jezik", "Bosnia and Herzegovina"),
        ("ca-ES", "ca", "cat", "Catalan", "Català", "Spain"),
        ("ce-RU", "ce", "che", "Chechen", "Нохчийн мотт", "Russian Federation"),
        ("ch-GU", "ch", "cha", "Chamorro", "Chamorru", "Guam"),
        ("co-FR", "co", "cos", "Corsican", "Corsu", "France"),
        ("cr-CA", "cr", "cre", "Cree", "ᓀᐦᐃᔭᐍᐏᐣ", "Canada"),
        ("cs-CZ", "cs", "ces", "Czech", "Český", "Czechia"),
        ("cu-BG", "cu", "chu", "Church Slavic", "Словѣньскъ", "Bulgaria"),
        ("cv-RU", "cv", "chv", "Chuvash", "Чӑвашла", "Russian Federation"),
        ("cy-GB", "cy", "cym", "Welsh", "Cymraeg", "United Kingdom"),
        ("da-DK", "da", "dan", "Danish", "Dansk", "Denmark"),
        ("de-DE", "de", "deu", "German", "Deutsch (Deutschland)", "Germany"),  # Default
        ("de-AT", "de", "deu", "German (Austria)", "Deutsch (Österreich)", "Austria"),
        (
            "de-CH",
            "de",
            "deu",
            "German (Switzerland)",
            "Deutsch (Schweiz)",
            "Switzerland",
        ),
        (
            "de-LI",
            "de",
            "deu",
            "German (Liechtenstein)",
            "Deutsch (Liechtenstein)",
            "Liechtenstein",
        ),
        (
            "en-US",
            "en",
            "eng",
            "English",
            "English (United States)",
            "United States of America",
        ),  # Default
        (
            "en-GB",
            "en",
            "eng",
            "English (United Kingdom)",
            "English (United Kingdom)",
            "United Kingdom",
        ),
        (
            "en-AU",
            "en",
            "eng",
            "English (Australia)",
            "English (Australia)",
            "Australia",
        ),
        ("en-CA", "en", "eng", "English (Canada)", "English (Canada)", "Canada"),
        ("dv-MV", "dv", "div", "Divehi", "ދިވެހި (Divehi)", "Maldives"),
        ("dz-BT", "dz", "dzo", "Dzongkha", "རྫོང་ཁ་ (Dzongkha)", "Bhutan"),
        ("ee-GH", "ee", "ewe", "Ewe", "Ɛʋɛgbɛ", "Ghana"),
        ("el-GR", "el", "ell", "Greek", "Ελληνικά", "Greece"),
        ("eo-001", "eo", "epo", "Esperanto", "Esperanto", "International"),
        ("es-ES", "es", "spa", "Spanish", "Español (España)", "Spain"),  # Default
        (
            "es-419",
            "es",
            "spa",
            "Spanish (Latin America)",
            "Español (Latinoamérica)",
            "Latin America",
        ),
        ("es-MX", "es", "spa", "Spanish (Mexico)", "Español (México)", "Mexico"),
        (
            "es-US",
            "es",
            "spa",
            "Spanish (United States)",
            "Español (Estados Unidos)",
            "United States of America",
        ),
        ("et-EE", "et", "est", "Estonian", "Eesti keel", "Estonia"),
        ("eu-ES", "eu", "eus", "Basque", "Euskara", "Spain"),
        ("fa-IR", "fa", "fas", "Persian", "فارسی", "Iran"),
        ("ff-SN", "ff", "ful", "Fula", "Fulfulde", "Senegal"),
        ("fi-FI", "fi", "fin", "Finnish", "Suomi", "Finland"),
        ("fil-PH", "fil", "phi", "Filipino", "Filipino", "Philippines"),
        ("fj-FJ", "fj", "fij", "Fijian", "Na Vosa Vakaviti", "Fiji"),
        ("fo-FO", "fo", "fao", "Faroese", "Føroyskt", "Faroe Islands"),
        ("fr-FR", "fr", "fra", "French", "Français (France)", "France"),  # Default
        ("fr-CA", "fr", "fra", "French (Canada)", "Français (Canada)", "Canada"),
        ("fy-NL", "fy", "fry", "Western Frisian", "Frysk", "Netherlands"),
        ("ga-IE", "ga", "gle", "Irish", "Gaeilge", "Ireland"),
        ("gd-GB", "gd", "gla", "Scottish Gaelic", "Gàidhlig", "United Kingdom"),
        ("gl-ES", "gl", "glg", "Galician", "Galego", "Spain"),
        ("gn-PY", "gn", "grn", "Guaraní", "Avañe'ẽ", "Paraguay"),
        ("gu-IN", "gu", "guj", "Gujarati", "ગુજરાત (Gujarātī)", "India"),
        ("gv-IM", "gv", "glv", "Manx", "Gaelg", "Isle of Man"),
        ("ha-NG", "ha", "hau", "Hausa", "هَوُس (Hausa)َ", "Nigeria"),
        ("he-IL", "he", "heb", "Hebrew", "עברית", "Israel"),
        ("hi-IN", "hi", "hin", "Hindi", "मानक हिन्दी", "India"),
        ("ho-PG", "ho", "hmo", "Hiri Motu", "Hiri Motu", "Papua New Guinea"),
        ("hr-HR", "hr", "hrv", "Croatian", "Hrvatski", "Croatia"),
        ("ht-HT", "ht", "hat", "Haitian Creole", "Kreyòl Ayisyen", "Haiti"),
        ("hu-HU", "hu", "hun", "Hungarian", "Magyar", "Hungary"),
        ("hy-AM", "hy", "hye", "Armenian", "հայերեն (Hayeren)", "Armenia"),
        ("hz-NA", "hz", "her", "Herero", "Otjiherero", "Namibia"),
        ("ia-001", "ia", "ina", "Interlingua", "Interlingua", "International"),
        ("id-ID", "id", "ind", "Indonesian", "Bahasa (Indonesia)", "Indonesia"),
        ("ie-001", "ie", "ile", "Interlingue", "Interlingue", "International"),
        ("ig-NG", "ig", "ibo", "Igbo", "Igbo", "Nigeria"),
        ("ii-CN", "ii", "iii", "Nuosu", "Nuosu, ꆈꌠꉙ", "China"),
        ("ik-US", "ik", "ipk", "Inupiaq", "Iñupiaq", "United States of America"),
        ("io-001", "io", "ido", "Ido", "Ido", "International"),
        ("is-IS", "is", "isl", "Icelandic", "Íslenska", "Iceland"),
        ("it-IT", "it", "ita", "Italian", "Italiano", "Italy"),
        ("iu-CA", "iu", "iku", "Inuktitut", "ᐃᓄᒃᑎᑐᑦ (Inuktitut)", "Canada"),
        ("ja-JP", "ja", "jpn", "Japanese", "日本語", "Japan"),
        ("jv-ID", "jv", "jav", "Javanese", "Jawa", "Indonesia"),
        ("ka-GE", "ka", "kat", "Georgian", "ქართული", "Georgia"),
        ("kg-CD", "kg", "kon", "Kongo", "KiKongo", "Congo, DR"),
        ("ki-KE", "ki", "kik", "Kikuyu", "Gĩgĩkũyũ", "Kenya"),
        ("kj-NA", "kj", "kua", "Kwanyama", "Oshikwanyama", "Namibia"),
        ("kk-KZ", "kk", "kaz", "Kazakh", "қазақ тілі", "Kazakhstan"),
        ("kl-GL", "kl", "kal", "Kalaallisut", "Kalaallisut", "Greenland"),
        ("km-KH", "km", "khm", "Khmer", "ខ្មែរ (Khmer)", "Cambodia"),
        ("kn-IN", "kn", "kan", "Kannada", "ಕನ್ನಡ", "India"),
        ("ko-KR", "ko", "kor", "Korean", "한국말", "Korea"),
        ("kr-NG", "kr", "kau", "Kanuri", "Kanuri", "Nigeria"),
        ("ks-IN", "ks", "kas", "Kashmiri", "کٲشُر (Koshur)", "India"),
        ("ku-TR", "ku", "kur", "Kurdish", "Kurdî", "Turkiye"),
        ("kv-RU", "kv", "kom", "Komi", "коми кыв", "Russian Federation"),
        ("kw-GB", "kw", "cor", "Cornish", "Kernewek", "United Kingdom"),
        ("ky-KG", "ky", "kir", "Kyrgyz", "кыргыз тили", "Kyrgyzstan"),
        ("la-VA", "la", "lat", "Latin", "Latina", "Holy See"),
        ("lb-LU", "lb", "ltz", "Luxembourgish", "Lëtzebuergesch", "Luxembourg"),
        ("lg-UG", "lg", "lug", "Ganda", "Luganda", "Uganda"),
        ("li-NL", "li", "lim", "Limburgan", "Lèmburgs", "Netherlands"),
        ("ln-CD", "ln", "lin", "Lingala", "Lingala", "Congo, DR"),
        ("lo-LA", "lo", "lao", "Lao", "ພາສາລາວ", "Lao People's Democratic Republic"),
        ("lt-LT", "lt", "lit", "Lithuanian", "Lietuviškai", "Lithuania"),
        ("lu-CD", "lu", "lub", "Luba-Katanga", "Kiluba", "Congo, DR"),
        ("lv-LV", "lv", "lav", "Latvian", "Latviešu", "Latvia"),
        ("mg-MG", "mg", "mlg", "Malagasy", "Fiteny", "Madagascar"),
        ("mh-MH", "mh", "mah", "Marshallese", "Kajin", "Marshall Islands"),
        ("mi-NZ", "mi", "mri", "Maori", "te reo Māori", "New Zealand"),
        ("mk-MK", "mk", "mkd", "Macedonian", "македонски", "North Macedonia"),
        ("ml-IN", "ml", "mal", "Malayalam", "മലയാളം", "India"),
        ("mn-MN", "mn", "mon", "Mongolian", "Монгол", "Mongolia"),
        ("mo-MD", "mo", "mol", "Moldavian", "молдовеняскэ", "Moldova"),
        ("mr-IN", "mr", "mar", "Marathi", "मराठी", "India"),
        ("ms-MY", "ms", "msa", "Malay", "Bahasa Melayu", "Malaysia"),
        ("mt-MT", "mt", "mlt", "Maltese", "Malti", "Malta"),
        ("my-MM", "my", "mya", "Burmese", "မြန်မာ", "Myanmar"),
        ("na-NR", "na", "nau", "Nauru", "Naoero", "Nauru"),
        ("nb-NO", "nb", "nob", "Norwegian Bokmål", "Norsk bokmål", "Norway"),
        ("nd-ZW", "nd", "nde", "North Ndebele", "isiNdebele", "Zimbabwe"),
        ("ne-NP", "ne", "nep", "Nepali", "नेपाली", "Nepal"),
        ("ng-NA", "ng", "ndo", "Ndonga", "Oshindonga", "Namibia"),
        ("nl-NL", "nl", "nld", "Dutch", "Nederlands", "Netherlands"),
        ("nn-NO", "nn", "nno", "Norwegian Nynorsk", "Norsk nynorsk", "Norway"),
        ("no-NO", "no", "nor", "Norwegian", "Norsk", "Norway"),
        ("nr-ZA", "nr", "nbl", "South Ndebele", "IsiNdebele", "South Africa"),
        ("nv-US", "nv", "nav", "Navajo", "Diné Bizaad", "United States of America"),
        ("ny-MW", "ny", "nya", "Chichewa", "Chicheŵa", "Malawi"),
        ("oc-FR", "oc", "oci", "Occitan", "Occitan", "France"),
        ("oj-CA", "oj", "oji", "Ojibwe", "ᐊᓂᔑᓈᐯᒧᐎᓐ", "Canada"),
        ("om-ET", "om", "orm", "Oromo", "Afaan Oromoo", "Ethiopia"),
        ("or-IN", "or", "ori", "Oriya", "ଓଡ଼ିଆ", "India"),
        ("os-RU", "os", "oss", "Ossetian", "ирон æвзаг", "Russian Federation"),
        ("pa-IN", "pa", "pan", "Panjabi", "ਪੰਜਾਬੀ", "India"),
        ("pi-IN", "pi", "pli", "Pali", "पालि", "India"),
        ("pl-PL", "pl", "pol", "Polish", "Polski", "Poland"),
        ("ps-AF", "ps", "pus", "Pashto", "پښ토", "Afghanistan"),
        (
            "pt-PT",
            "pt",
            "por",
            "Portuguese",
            "Português (Portugal)",
            "Portugal",
        ),  # Default
        ("pt-BR", "pt", "por", "Portuguese (Brazil)", "Português (Brasil)", "Brazil"),
        ("qu-PE", "qu", "que", "Quechua", "Runa Simi", "Peru"),
        ("rc-RE", "rc", "rcf", "Réunion Creole", "Kréol rénioné", "Reunion"),
        ("rm-CH", "rm", "roh", "Romansh", "Rumantsch", "Switzerland"),
        ("rn-BI", "rn", "run", "Ikirundi", "Burundi", "Burundi"),
        ("ro-RO", "ro", "ron", "Romanian", "Româneşte", "Romania"),
        ("ru-RU", "ru", "rus", "Russian", "русский", "Russian Federation"),
        ("rw-RW", "rw", "kin", "Kinyarwanda", "Ikinyarwanda", "Rwanda"),
        ("sa-IN", "sa", "san", "Sanskrit", "संस्कृतम्", "India"),
        ("sc-IT", "sc", "srd", "Sardinian", "Sardu", "Italy"),
        ("sd-PK", "sd", "snd", "Sindhi", "سنڌي", "Pakistan"),
        ("se-NO", "se", "sme", "Northern Sami", "Sámegiella", "Norway"),
        ("sg-CF", "sg", "sag", "Sango", "Sango", "Central African Republic"),
        (
            "sh-BA",
            "sh",
            "hbs",
            "Serbo-Croatian",
            "Srpskohrvatski",
            "Bosnia and Herzegovina",
        ),
        ("si-LK", "si", "sin", "Sinhala", "සිංහල", "Sri Lanka"),
        ("sk-SK", "sk", "slk", "Slovak", "Slovenčina", "Slovakia"),
        ("sl-SI", "sl", "slv", "Slovenian", "Slovenščina", "Slovenia"),
        ("sm-WS", "sm", "smo", "Samoan", "Gagana Samoa", "Samoa"),
        ("sn-ZW", "sn", "sna", "Shona", "Chishona", "Zimbabwe"),
        ("so-SO", "so", "som", "Somali", "Af-Soomaali", "Somalia"),
        ("sq-AL", "sq", "sqi", "Albanian", "Shqip", "Albania"),
        ("sr-RS", "sr", "srp", "Serbian", "српски", "Serbia"),
        ("ss-SZ", "ss", "ssw", "Swati", "siSwati", "Eswatini"),
        ("st-LS", "st", "sot", "Southern Sotho", "Sesotho", "Lesotho"),
        ("su-ID", "su", "sun", "Sundanese", "Basa Sunda", "Indonesia"),
        ("sv-SE", "sv", "swe", "Swedish", "Svenska", "Sweden"),
        ("sw-KE", "sw", "swa", "Swahili", "Kiswahili", "Kenya"),
        ("ta-IN", "ta", "tam", "Tamil", "தமிழ்", "India"),
        ("te-IN", "te", "tel", "తెలుగు", "India"),
        ("tg-TJ", "tg", "tgk", "Tajik", "тоҷикӣ (Tojiki)", "Tajikistan"),
        ("th-TH", "th", "tha", "Thai", "ภาษาไทย‎", "Thailand"),
        ("ti-ER", "ti", "tir", "Tigrinya", "ትግርኛ (Tigrinya)", "Eritrea"),
        ("tk-TM", "tk", "tuk", "Turkmen", "Türkmençe", "Turkmenistan"),
        ("tl-PH", "tl", "tgl", "Tagalog", "Tagalog", "Philippines"),
        ("tn-BW", "tn", "tsn", "Tswana", "Setswana", "Botswana"),
        ("to-TO", "to", "ton", "Tongan", "Faka Tonga", "Tonga"),
        ("tr-TR", "tr", "tur", "Turkish", "Türkçe", "Turkiye"),
        ("ts-ZA", "ts", "tso", "Tsonga", "Xitsonga", "South Africa"),
        ("tt-RU", "tt", "tat", "Tatar", "татарча", "Russian Federation"),
        ("tw-GH", "tw", "twi", "Twi", "Twi", "Ghana"),
        ("ty-PF", "ty", "tah", "Tahitian", "Reo Tahiti", "French Polynesia"),
        (
            "zh-CN",
            "zh",
            "zho",
            "Chinese (Simplified)",
            "中文 (中国大陆)",
            "China",
        ),  # Default
        ("zh-TW", "zh", "zho", "Chinese (Traditional)", "中文 (臺灣)", "Taiwan"),
        ("zh-HK", "zh", "zho", "Chinese (Hong Kong)", "中文 (香港)", "Hong Kong"),
        ("ug-CN", "ug", "uig", "Uyghur", "ئۇيغۇرچە (Uyghurche)", "China"),
        ("uk-UA", "uk", "ukr", "Ukrainian", "українська", "Ukraine"),
        ("ur-PK", "ur", "urd", "Urdu", "اردو (Urdū)", "Pakistan"),
        ("uz-UZ", "uz", "uzb", "Uzbek", "O'zbek", "Uzbekistan"),
        ("ve-ZA", "ve", "ven", "Venda", "Tshivenḓa", "South Africa"),
        ("vi-VN", "vi", "vie", "Vietnamese", "Việt", "Viet Nam"),
        ("vo-001", "vo", "vol", "Volapük", "Volapük", "International"),
        ("wa-BE", "wa", "wln", "Walloon", "Walon", "Belgium"),
        ("wo-SN", "wo", "wol", "Wolof", "Wolof", "Senegal"),
        ("xh-ZA", "xh", "xho", "Xhosa", "isiXhosa", "South Africa"),
        ("yi-001", "yi", "yid", "Yiddish", "ייִדיש", "International"),
        ("yo-NG", "yo", "yor", "Yoruba", "Yorùbá", "Nigeria"),
        ("za-CN", "za", "zha", "Zhuang", "Saɯ cueŋƅ", "China"),
        ("zu-ZA", "zu", "zul", "Zulu", "isiZulu", "South Africa"),
    ]
    return language_matrix_references


_matrix = get_language_matrix()

# Map BCP 47 code (e.g., "en-US") to its long name ("English (United States)")
_bcp47_map = {entry[0]: entry[4] for entry in _matrix}

# Map ISO 639-2/B code (e.g., "eng") to its long name ("English")
# We use the designated default for this mapping
_iso639_2_defaults = {
    "deu": "de-DE",
    "eng": "en-US",
    "spa": "es-ES",
    "fra": "fr-FR",
    "por": "pt-PT",
    "zho": "zh-CN",
}
_iso639_2_map = {
    code: _bcp47_map.get(bcp47) for code, bcp47 in _iso639_2_defaults.items()
}

# Map for 2-letter codes to their designated default BCP 47 code
_iso639_1_defaults = {
    "de": "de-DE",
    "en": "en-US",
    "es": "es-ES",
    "fr": "fr-FR",
    "pt": "pt-PT",
    "zh": "zh-CN",
}


def get_long_language_name(code: str) -> Optional[str]:
    """
    Looks up the long, descriptive language name from various codes.
    - BCP 47 (e.g., "en-US")
    - ISO 639-2/B (e.g., "eng")
    - ISO 639-1 (e.g., "en"), respecting designated defaults.
    """
    if not code:
        return None

    # 1. Direct BCP 47 match (most specific)
    if "-" in code and code in _bcp47_map:
        return _bcp47_map[code]

    # 2. ISO 639-2/B (3-letter) match
    if len(code) == 3 and code in _iso639_2_map:
        return _iso639_2_map[code]

    # 3. ISO 639-1 (2-letter) match using defaults
    if len(code) == 2 and code in _iso639_1_defaults:
        default_bcp47 = _iso639_1_defaults[code]
        return _bcp47_map.get(default_bcp47)

    # Fallback for 3-letter codes without a specified default
    if len(code) == 3:
        for entry in _matrix:
            if entry[2] == code:
                return entry[4]

    # Fallback for 2-letter codes without a specified default
    if len(code) == 2:
        for entry in _matrix:
            if entry[1] == code:
                return entry[4]

    return None



# metaspector/format_handlers/__init__.py
# !/usr/bin/env python3

# This file makes the 'format_handlers' directory a Python package.
# You can optionally import specific format parsers here for convenience,
# but for now, it can remain empty or contain a simple docstring.

"""
This package contains modules for parsing different media file formats.
Each submodule (e.g., mp4.py, flac.py) handles the parsing logic for a specific format.
"""
